# Copyright 2025 qBraid
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Profile module for defining and registering QIR translation profiles.

This package exposes base and adaptive profiles, along with shared interfaces
and registry mechanisms to manage profile capabilities and restrictions.
"""

from .core import (
    AdaptiveProfile,
    BaseProfile,
    Profile,
    ProfileCapabilities,
    ProfileRegistry,
    ProfileRestrictions,
)

__all__ = [
    "ProfileCapabilities",
    "ProfileRestrictions",
    "Profile",
    "BaseProfile",
    "AdaptiveProfile",
    "ProfileRegistry",
]
