# Copyright 2025 qBraid
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Abstract base class for QIR module.

"""

from abc import ABC, abstractmethod
from typing import Any


class QIRModule(ABC):
    """Abstract base class for QIR modules."""

    def __init__(self, program: Any) -> None:
        self.program = program

    @property
    @abstractmethod
    def num_qubits(self) -> int:
        """Returns the number of qubits in the program."""

    @property
    @abstractmethod
    def num_clbits(self) -> int:
        """Returns the number of classical bits in the program."""
